#!/usr/bin/env python

'''
'''
#----------------------------------------------------------------------------------
import os, sys, datetime, re, json, shutil, glob, traceback
import pandas as pd
import numpy as np
sys.path.append("/opt/utils")
from colabexts import utils as colabexts_utils
from collections import defaultdict 
from mangorest import mango
from inspect import isfunction

import logging

# Remove all handlers associated with the root logger object.
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)
logging.basicConfig( level=logging.INFO,
        format='%(levelname)s:%(name)s %(asctime)s %(filename)s:%(lineno)s:%(funcName)s: %(message)s',
        #format='%(asctime)s %(name)s %(levelname)s: %(message)s',
        handlers=[ logging.FileHandler("/tmp/app.log"), logging.StreamHandler()],
)
logger = logging.getLogger( "genai_utils" )


class myjson:
    def __init__(self, db="default", base=os.path.expanduser("~/myjson")):
        self.base = os.path.expanduser(base)
        self.db = db
        self.createdb( db )

    def createdb(self, db):
        db = self.base + "/" + db
        os.makedirs(db, exist_ok=True)
        self.db = db
        logger.debug(f"Current DB {self.db}")
        #traceback.print_stack()
        return self.db

    def changedb(self, db):
        createdb(db)

    def deletedb(self, db):
        if (db == "default"):
            logger.warning(f"Cannot Delete default DB")
            return

        db = self.base + "/" + db
        logger.info(f"Deleting DB {db}: {db}")
        shutil.rmtree(db)
        if self.db == db:
            createdb("default")

        return self.db
        

    def listDbs(self):
        dbs = [os.path.basename(c[0:-1]) for c in glob.glob(f"{self.base}/*/", recursive=True)]
        return dbs

    def listTables(self):
        dbs = [os.path.basename(c[0:-1]) for c in glob.glob(f"{self.db}/*/", recursive=True)]
        return dbs

    def create_table(self, table):
        tab = self.db + "/" + table 
        os.makedirs(tab, exist_ok=True)
        return tab

    def delete_table(self, table):
        tab = self.db + "/" + table 
        if ( os.path.exists(tab)):
            shutil.rmtree(tab)
        return tab

    def _update_info(self, df, table, max_id):
        if ( len(df) <= 0):
            return
            
        info_file = f"{self.db}/{table}/info.js"
        infoj = dict(columns=[c for c in df.columns], maxid=max_id, nrows=len(df))
        open(info_file, "w").write(json.dumps(infoj, indent=2))

        #infoj = colabexts_utils.parsej(open(info_file).read())
        return infoj
        
    def fromDataFrame(self, df, table, delete=True):
        if (delete):
           self.delete_table(table) 
        tab = self.create_table(table)
        for rowid, r in df.iterrows():
            d = r.to_dict()
            d['rowid'] = rowid
            open(f"{tab}/{rowid}.json", "w").write( json.dumps(d, indent=2))
        
    def get(self, table, rowid=None ):
        if ( not rowid):
            return self.read(table)

        f = f"{self.db}/{table}/{rowid}.json"
        if ( not os.path.isfile(f)):
            return None
        r = open(f).read()
        a = colabexts_utils.parsej(r)
        return a;

    def read(self, table, nrows=1024*1024, filter=None , **kwargs):
        rows, cols, max_id =[], {}, 0

        dir = f"{self.db}/{table}/*.json"

        #logger.debug(f"Reading from {dir}")

        sortk =  lambda x: int(os.path.basename(x).split('.')[0])
        if ( nrows and nrows < 0):
            nrows = -nrows
            files = sorted(glob.glob(dir), key=sortk , reverse=True)
        else:
            files = sorted(glob.glob(dir), key=sortk )

        for f in files:
            try:
                r = open(f).read()
                a = colabexts_utils.parsej(r)
                #a = json.loads(r)
                if ("rowid" not in a):
                    logger.error(f"ROWID IS MISSING in {f}")
                    #a['rowid'] = os.path.basename(f)[:-5]
                    continue; 
                    
                max_id = a['rowid']

                if ( filter and type(filter) == dict):
                    c = all(a.get(k, None) == v for k, v in filter.items())
                    if (not c):
                        continue
                elif ( filter and isfunction(filter) ):
                    c= filter(a)
                    if (not c):
                        continue

                cols.update(a)
                rows.append(a)
            except Exception as e:
                logger.error(e)
                out = ""
                for i, l in enumerate(r.split("\n")[:100]):
                    out += f'{i+1:3d}: {l}\n'
                logger.error(f"Error while parsing {f} \n\n{out}")
            if ( len(rows) >= nrows):
                break;
                
        df = pd.DataFrame(rows) 
        if ( not filter):
            self._update_info (df, table, max_id,)
        return df

    def delete(self, table, rowid, **kwargs):
        print(f'''
        **** DELETING {table} {rowid}
        **** 
        ****
        ''')
        f = f"{self.db}/{table}/{rowid}.json"
        if ( not os.path.isfile(f)):
            return 0
        os.rename(f, f+".deleted")
        pass;


    def update(self, table, data):
        ret = data
        if type(data) == str:
            ret = colabexts_utils.parsej(data)
        elif type(data) == pd.Series:
            ret = data.to_dict()
            
        tab = self.db + "/" + table 

        rowid = ret.get("rowid", None)
        if ( rowid is None or not str(rowid) or int(rowid) < 0 ):
            rowid = max([-1] +[int(os.path.basename(c[:-5])) for c in glob.glob(f"{tab}/*.json")])
            rowid = rowid + 1
            logger.info(f"Update called with no rowid: Assuming insert => newId= {rowid}")
            ret['rowid'] = rowid
            
        rowid = ret["rowid"]
        row = f"{tab}/{rowid}.json"
        
        if ( os.path.exists(row)):
            old = colabexts_utils.parsej((open( row, "r")).read())
            old.update(ret)
            ret = old
            
        open(f"{tab}/{rowid}.json", "w").write( json.dumps(ret, cls=mango.myEncoder, indent=2))

        return ret
        
#db = myjson()



#!/usr/bin/env python 

import sys, os, torch, logging,datetime, httpx,re, hashlib, json, base64
from ollama import Client
import ollama
from mangorest.mango import webapi
from openai import OpenAI

logger = logging.getLogger( "genai_utils" )

BASE = os.path.expanduser("~/data/gpt/")

#------------------------------------------INITIALIZE LLM Stuff -------------------- 
device = "cpu"
if (torch.cuda.is_available() ):
    device = "cuda"
elif torch.backends.mps.is_available():
    device = "mps"

OLLAMA_HOST= 'http://127.0.0.1:11434/v1'
OPENAI_KEY = "NO KEY"

sys.path.append(os.path.expanduser("~/.django") )
if (os.path.exists(os.path.expanduser("~/.django/my_config.py"))):
    import my_config
    try:
        from my_config import OLLAMA_HOST
        OPENAI_KEY=my_config.OPENAI_KEY
    except:
        pass
#------------------------------------------INITIALIZE the DB-------------------- 
from gpt.jsondb import myjson

MYDB = myjson(base=BASE, db='INDICES')
INDEX_TABLE = "indexes"
#------------------------------------------------------------------------------------------ 
@webapi("/gpt/getKBs/")
def getKBs(request=None, **kwargs):
    df = MYDB.read( INDEX_TABLE, nrows=-500)
    ret = {
        'columns': [c for c in df.columns],
        'values' : df.values.tolist()
    }
    return ret
#------------------------------------------------------------------------------------------ 
def isValid(file, **kwargs):
    return True

@webapi("/gpt/deleteKB/")
def deleteKB(request=None, rowid="-1", **kwargs):
    print(locals())
    MYDB.delete(INDEX_TABLE, rowid)
    return getKBs()

@webapi("/gpt/uploadFiles/")
def uploadFiles(request=None, name="test", **kwargs):
    ret = ""
    source_folder=kwargs.get("source_folder", "")
    if ( request):
        for f in request.FILES.getlist('file'):
            content = f.read()
            if ( not source_folder):
                basedir =  f"{BASE}/{name}-files/"
            else:
                basedir =  f"{BASE}/{source_folder}/"
            os.makedirs(basedir, exist_ok=True)
            filename = f"{basedir}/{str(f)}"

            ret += f"{filename} => uploading \n"
            if ( isValid(name, **kwargs)):
                with open(filename, "wb") as f:
                    f.write(content)
            
    return ret

def indexFiles(request, **kwargs):
    cmd = ""
    if (  len(request.FILES) > 0):
        v = [kwargs.get(k, "") for k in "source_folder es_url es_user es_pass index_name model".split()]
        v[0] = f"{BASE}/{v[0]}"     # Application specif folder

        vision_index=kwargs.get("vision_index", "")
        if ( not vision_index ):
            cmd = f"gpt/db_elastic.py -p {v[0]} -e {v[1]} -u '{v[2]}' -w '{v[3]}' -i {v[4]} -m {v[5]}"

        if ( vision_index ):
            cmd = f"gpt/db_vision.py -p {v[0]} "
            pass

    if (cmd):
        logger.info("Executing " + cmd)
        os.system(cmd + " &")


@webapi("/gpt/createUpdateKB/")
def createUpdateKB(request=None,  **kwargs):
    MYDB.create_table(INDEX_TABLE)
    df = MYDB.update( INDEX_TABLE, kwargs)

    # index_name must be sent
    index_name = kwargs.get("index_name")

    ret = uploadFiles(request, index_name, **kwargs)            
    indexFiles(request, **kwargs)

    return getKBs()

@webapi("/gpt/searchKB/")
def searchKB(request=None,  **kwargs):
    import db_elastic, db_vision
    keys = "source_folder query vision_index model".split()
    [path, query, v,model] = [kwargs.get(k, "") for k in keys]
    path = f"{BASE}/{path}"     # Application specif folder

    if ( v ):
        resp = db_vision.queryVision(query, path, model=model)
        ret  = [dict(page_content=resp, metadata="None")]
    else:
        ret = db_elastic.esSearchIndex(request, **kwargs)

    return ret
    pass;




#!/usr/bin/env python 

'''
RUN as 'python -m genai_utils.db_elastic -p "
'''

import os, sys, logging, argparse, glob, hashlib

from langchain_ollama import OllamaEmbeddings
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from langchain_elasticsearch import (
    BM25Strategy,
    DenseVectorStrategy,
    ElasticsearchStore,
)

from elasticsearch import Elasticsearch
from mangorest.mango import webapi
from genai_utils import pdf_parser
from genai_utils import extract_docs

logger = logging.getLogger( "gpt" )

ES_URL, ES_USER, ES_PW  = "http://localhost:9200", "elastic", "elastic"

from genai_utils import config

ES_CNX= dict(es_url= ES_URL, es_user= ES_USER, es_password=ES_PW)


_ES_STARTEGIES = {
    "hnsw":     DenseVectorStrategy(), 
    "bm25":     BM25Strategy(),
    "hybrid":   DenseVectorStrategy(hybrid=True, rrf=False),
    "sparse":   None,
    "exact":    None,
}
# ---------------------------------------------------------------------------------------
def esDeleteIndex(index="test", es_url=ES_URL, es_user=ES_USER, es_pass= ES_PW, **kwargs):
    esclient = Elasticsearch(es_url, basic_auth = (es_user, es_pass))
    esclient.info()
    try:
        esclient.indices.delete(index=index)
    except:
        pass
# ---------------------------------------------------------------------------------------
def esCreateIndex(index="test", es_url=ES_URL, es_user=ES_USER, es_pass= ES_PW, **kwargs):
    esclient = Elasticsearch(es_url, basic_auth = (es_user, es_pass))
    esclient.indices.create(index=index)
# ---------------------------------------------------------------------------------------
def esCountIndex(index="test", es_url=ES_URL, es_user=ES_USER, es_pass= ES_PW, **kwargs):
    doc_count = 0
    try:
        esclient = Elasticsearch(es_url, basic_auth = (es_user, es_pass))
        doc_count = esclient.count(index=index)
    except:
        pass
    print(f"Total documents in index '{index}': {doc_count}")
    return doc_count
# ---------------------------------------------------------------------------------------
def getEmbedding(model="all-minilm:L6-v2", base_url = "http://127.0.0.1:11434/"):
    e = OllamaEmbeddings( model = model, base_url =base_url )
    return e
# ---------------------------------------------------------------------------------------
def getbyID( index="test",id="", es_url=ES_URL, es_user=ES_USER, es_pass= ES_PW, **kwargs):
    es_cnx = dict(es_url= es_url, es_user=es_user, es_password=es_pass)
    doc = ElasticsearchStore.get_by_ids(ids=[id]
            **es_cnx,
            index_name=index)
    return doc
# ---------------------------------------------------------------------------------------
def add_to_es( docs: list[Document], es_cnx: dict, index: str, embed, strategy= "hnsw" ):
    strat = _ES_STARTEGIES[strategy]
    vectorstore = None
    for i in range(0, len(docs), 20000):
        docsWithID = docs[i : min(i + 20000, len(docs))]
        for d in docsWithID:
            h = hashlib.md5(d.page_content.encode())
            d.id = h.hexdigest()
        
        vectorstore = ElasticsearchStore.from_documents(
            documents=docsWithID,
            embedding=embed,
            **es_cnx,
            index_name=index,
            bulk_kwargs={
                "chunk_size": 100,
            },
            strategy=strat,
        )
    return vectorstore

# ---------------------------------------------------------------------------------------
def es_retriever( es_cnx: dict, index: str, embed, strategy="hnsw", k= 10 ):
    strat = _ES_STARTEGIES[strategy]

    v = ElasticsearchStore( **es_cnx, embedding=embed, index_name=index, strategy=strat)
    return v.as_retriever(search_kwargs={"k": k})

def esVectorSearch( retreiver, q, k=10):
        ret = retreiver.as_retriever(search_kwargs={"k": k}).invoke(q)
        
        h = {r.page_content:r for r in ret}
        if len(h) != len(ret):
            ret = [v for v in h.values()]
            
        return ret

@webapi("/gpt/esSearchIndex/")
def esSearchIndex(request, index, query, model="all-minilm:L6-v2", user="", es_url=ES_URL, 
                    es_user=ES_URL, es_pass=ES_PW, k=10, rank=1, **kwargs):

    #print(f"\n{locals()}\n")
        
    if (not es_url):
        es = dict(es_url= ES_URL, es_user=ES_USER, es_password=ES_PW)
    else:
        es = dict(es_url= es_url, es_user=es_user, es_password=es_pass)

    #model = "llama3.2" #lets force the embedding for now
    embed = getEmbedding(model=model) 

    
    #if not os.path.exists(os.path.expanduser("~/.cache/RERANKER/")):
    #    print(f"**** Ranker cache does not exist ****")
    #    return ret
    if ( rank):
        v = es_retriever(es, index=index, embed=embed, k=k*2)
        docs = v.invoke(query)
        if (len(docs)):
            ranked = rerank( query, docs)
            docs = [Document(page_content=r['text'], metadata=r['metadata']) for r in ranked[0:k]]
    else:
        v = es_retriever(es, index=index, embed=embed, k=k)
        docs = v.invoke(query)

    h = {r.page_content: r for r in docs}
    if len(h) != len(docs):
        docs = [v for v in h.values()]
    
    ret = []
    for d in docs:
        ret.append(dict(page_content=d.page_content, metadata=d.metadata))
    return ret

# ---------------------------------------------------------------------------------------
def format(d, show=1):
    from IPython.display import HTML
    m=d['metadata']
    page = m['page'] if "page" in m  else "?"
    html=f'''
<h3>Document, {page} : {m['source']} </h3> 

{d['page_content'].replace("\n", "<br>")}
<hr/>
'''
    if(show):
        display(HTML(html))
    return html

# ---------------------------------------------------------------------------------------
@webapi("/gpt/esTextSearch/")
def esTextSearch(query, k=10, index="test", es_url = ES_URL, es_user=ES_USER, es_pass= ES_PW):
    esclient = Elasticsearch(es_url, basic_auth = (es_user, es_pass))
    res = esclient.search(index=index,  q=query, size=k)

    ret = []
    for i,r in enumerate(res['hits']['hits']):
        pc = r['_source']['text']
        mt = r['_source']['metadata']
        ret.append(dict(page_content = pc, metadata=mt))
        #print(i, " ==>", )
    return ret
# ---------------------------------------------------------------------------------------
def rerank(q, ret):
    from flashrank import (Ranker, RerankRequest,)
    
    ranker = Ranker("ms-marco-MiniLM-L-12-v2", os.path.expanduser("~/.cache/RERANKER/"))
    rerankrequest = RerankRequest(
        query=q, passages=[{"text": d.page_content, "metadata": d.metadata} for d in ret]
    )
    reranked = ranker.rerank(rerankrequest)
    return reranked
# ---------------------------------------------------------------------------------------
# This is standing by itself - should be called by indexFromFolder
# can be multi tasked 
def loadES( model="all-minilm:L6-v2", index="", filename = "/Users/e346104/Desktop/data/LLM/sample.pdf",
           es_url=ES_URL , es_user=ES_USER, es_pass=ES_PW, docs=[] ):
    
    if(not docs and filename):
        docs = extract_docs.extractDocs(file=filename)
        
    if (not docs):
        return docs
    embed= getEmbedding(model)
    es = dict(es_url=es_url , es_user=es_user, es_password=es_pass)
    v = add_to_es(docs, es, index=index, embed=embed)

    return docs

# ---------------------------------------------------------------------------------------
MARKER_BASE=f"/tmp/gpt/"
def indexFromFolder(folder="", force=0, index="test", recurse=0, just_show=0,
                        es_url=ES_URL, es_user=ES_USER, es_pass= ES_PW, model="all-minilm:L6-v2"):
    folder = os.path.expanduser(folder) + "/**"
    files = [f for f in glob.glob(folder, recursive=recurse) if os.path.isfile(f)]

    logger.info(f"URL: {es_url} Folder: {folder}: found {len(files)} files.")

    iFiles = []
    for f in files:
        bn = os.path.basename(f)
        dn = os.path.dirname(f)
        marker = f"{MARKER_BASE}/{index}/{dn}/.{bn}.indexed"

        if f.startswith(".") or f.endswith(".indexed") or (os.path.exists(marker) and not force):
            print(f"Already in cache '{f}' ... ")
            continue;

        try:
            if ( not just_show):
                pass
                logger.info(f"Indexing '{f}' {es_url}")        
                loadES(model, index, f, es_url, es_user, es_pass)
                os.makedirs(os.path.dirname(marker, mode=0o777), exist_ok=True)
                open(marker, "w", mode=0o777).write("")
                iFiles.append(f)
            else:
                print(f"Not indexing '{f}'\n======================")
        except Exception as e:
            logger.error(f"{f} failed to index {e}\n================")
            pass

    esCountIndex(index=index, es_url=es_url, user=es_user, pw= es_pass)
    return iFiles
#-----------------------------------------------------------------------------------
sysargs=None
def addargs(argv=sys.argv):
    global sysargs
    p = argparse.ArgumentParser(f"{os.path.basename(argv[0])}:")
    p.add_argument('-p', '--path',   required=False, type=str, default='.', help="path to look for files")
    p.add_argument('-i', '--index',  type=str, required=True, help="Elastic Search index")
    p.add_argument('-m', '--model',  type=str, required=False, default="all-minilm:L6-v2", 
                    help="embedding model; defaults to local ollama model 'all-minilm:L6-v2' ")
    p.add_argument('-e', '--es_url', type=str, required=False, default=ES_URL,  help=f"elastic URL default: {ES_URL}")
    p.add_argument('-u', '--es_user',type=str, required=False, default=ES_USER, help=f"elastic user. default: {ES_USER}")
    p.add_argument('-w', '--es_pass',type=str, required=False, default=ES_PW,   help=f"elastic password. default: {ES_PW}")
    p.add_argument('-f', '--force',  required=False, default=False, action='store_true', 
                   help="force and reindex - the files indexed will be ignored otherwise")
    p.add_argument('-j', '--just' ,  required=False, default=False, action='store_true', help="Just show - do not index")
    p.add_argument('-r', '--recurse',required=False, default=False, action='store_true', help="Recurse through the folder")
    p.add_argument('-q', '--query',required=False, type=str, default="", 
                   help="Search for context - instead of indexing - this will search the index")
    p.add_argument('-d', '--delete',required=False, default=False, action='store_true', help="Delete Index")

    sysargs=p.parse_args(argv[1:])
    return sysargs

from colabexts import utils as colabexts_utils
if __name__ == '__main__' and not colabexts_utils.inJupyter():
    a = addargs()
    logger.info(f"Indexing  {sysargs}")

    if ( a.delete):
        print(f"Deleting the index: {a.index}")
        esDeleteIndex(index=a.index, es_url=a.es_url, es_user=a.es_user, es_pass=a.es_pass )
        marker = f"{MARKER_BASE}/{a.index}/"
        os.rmdir(marker)
    elif ( a.query):
        print(f"Searching for context: {a.query}")
        res0 = esSearchIndex(None, index=a.index, es_url=a.es_url, es_user=a.es_user, es_pass=a.es_pass,  query=a.query)
        res1 = esTextSearch(query=a.query, index=a.index, es_url=a.es_url, es_user=a.es_user, k=5 )

        for d in res0:
            s = f"src: {d['metadata']['source']}\n{d['page_content']}\n{'-'*80}\n"
            print(s)
        for d in res1:
            s = f"src: {d['metadata']['source']}\n{d['page_content']}\n{'-'*80}\n"
            print(s)
    else:
        indexFromFolder(folder=a.path, force=a.force, index=a.index, es_url=a.es_url, recurse=a.recurse,
                        es_user=a.es_user, es_pass= a.es_pass, model=a.model)

#    indexFromFolder(sys.argv[1])
# index, model = "test2", "all-minilm:L6-v2"
# index, model = "test3", "llama3.2:latest"

# esDeleteIndex(index)
# esCreateIndex(index)

# loadES(model, index);



#!/usr/bin/env python

import os,glob
from mangorest.mango import webapi
from ray import logger
from langchain_core.documents import Document

BASE = "~/data/gpt/"
from langchain_ollama import OllamaEmbeddings
from langchain_elasticsearch import (
    BM25Strategy,
    DenseVectorStrategy,
    ElasticsearchStore,
)
# ---------------------------------------------------------------------------------------
# Elastic search parameters
#
ES_URL, ES_USER, ES_PW  = "http://localhost:9200", "elastic", "elastic"
ES_CNX   = dict(es_url= ES_URL, es_user= ES_USER, es_password=ES_PW)
_ES_STARTEGIES = {
    "hnsw":     DenseVectorStrategy(), 
    "bm25":     BM25Strategy(),
    "hybrid":   DenseVectorStrategy(hybrid=True, rrf=False),
    "sparse":   None,
    "exact":    None,
}
# ---------------------------------------------------------------------------------------
# Embeddings - lets use local emebddings
#
def getEmbedding(model="all-minilm:L6-v2",base_url = "http://127.0.0.1:11434/"):
    e = OllamaEmbeddings( model = model, base_url =base_url )
    return e

class Elastic:
    def __init__(self, es_cnx=ES_CNX, index="sageai", strategy="hnsw", embed="all-minilm:L6-v2"):
        self.es_cnx, self.index, self. strategy, self.embed  = es_cnx, index, strategy, embed
        if type(self.embed) == str:
            self.embed = getEmbedding(embed)
            
        self.strategy = _ES_STARTEGIES[self.strategy]
    # ---------------------------------------------------------------------------------------
    def _getChunks(self, file=f"{BASE}HS4_SGS1_V1S2.pdf", chunk_overlap_ratio=0.2, chunk_size=8000):
        from rag_basic.process_files import multi_process_pdf,process_pdf_to_chunks
        if ( file.endswith(".pdf")):
            chunks = process_pdf_to_chunks(file)        
        elif( file.endswith(".txt")):
            with open(file, "r", encoding="utf-8", errors='ignore') as f:
                text = f.read()
                
            from langchain_text_splitters import RecursiveCharacterTextSplitter
            text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=chunk_size,
                chunk_overlap= int(chunk_overlap_ratio * chunk_size),
                add_start_index=True,
            )
            texts = text_splitter.split_text(text)
            chunks = []
            for t in texts:
                chunks.append(Document( page_content=t,  metadata={"source": file}))
        else:
            assert 0, "Unknown file type"
        return chunks
    # ---------------------------------------------------------------------------------------
    def indexdoc(self, file=f"{BASE}HS4_SGS1_V1S2.pdf", embed=None, strategy="hnsw"):
        file = os.path.expanduser(file)
        chunks = self._getChunks(file)

        for i in range(0, len(chunks), 20000):
            vectorstore = ElasticsearchStore.from_documents(
                documents= chunks[i : min(i + 20000, len(chunks))],
                embedding=self.embed,
                **self.es_cnx,
                index_name=self.index,
                bulk_kwargs={
                    "chunk_size": 100,
                },
                strategy=self.strategy,
            )
        
        print(f"Indexed {len(chunks)} chunks from {file}")
    # ---------------------------------------------------------------------------------------
    def es_retriever( self, k= 10 ):
        v = ElasticsearchStore( **self.es_cnx, embedding=self.embed, index_name=self.index, strategy=self.strategy)
        return v.as_retriever(search_kwargs={"k": k})

    # ---------------------------------------------------------------------------------------
    def search(self,query, k=10, rerank=1, **kwargs):
        docs = self.es_retriever().invoke(query)

        h = {r.page_content: r for r in docs}
        if len(h) != len(docs):
            docs = [v for v in h.values()]
        
        ret = []
        for d in docs:
            ret.append(dict(page_content=d.page_content, metadata=d.metadata))
        return ret

def indexFolder(folder=".", recursive=1):
    folder = os.path.expanduser((folder))
    logger.info(f"Indexing {folder}")
    se = Elastic()
    files = [f for f in glob.glob(folder, recursive=recursive) if os.path.isfile(f)]
    for f in files:
        print(f"{os.path.getsize(f)} : {f} ")
        se.indexdoc(f)
        
from colabexts import utils as colabexts_utils
if __name__ == '__main__' and not colabexts_utils.inJupyter():
    folder = "." if len(sys.args) <= 1 else sys.args[1]
    logger.info(f"Indexing  {folder}")
    indexFolder(folder)


def test():
    se = Elastic()
    docs = se.search("tell me about Ka-band - where do they play")
    return docs
    



#!/usr/bin/env python 

'''
python -m genai_utils.index_imges --directory </path/to/images>

'''
from ollama import generate

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
DEFAULT_MODEL = 'qwen2.5'

DEFAULT_SYSTEM_PROMPT = """
You are an image analyst who has been tasked with describing images.
Your goal is to enable retrieving the images based on the content of your descriptions.
"""

DEFAULT_USER_PROMPT = """
Describe this image.
Provide a thorough and detailed description, focusing on identifying and describing objects in the image.
Use denotative rather than connotative language. 
Read the text in the images and include it in the description without including the location of font name.
Do not explain how you are describing the image. Do not use any "I" statements.
"""

def describe_image(image_data, prompt=DEFAULT_USER_PROMPT, system=DEFAULT_SYSTEM_PROMPT, model=DEFAULT_MODEL):
    print(f"Describe image: {model}")
    result = generate(model=model, prompt=prompt, images=[image_data])
    return result['response']



def checkMatched(query, descritpion):
    prompt=f"""
You are a great critic who can rank query and descritpion on the scale of 1 to 5.
You will give 0 if the query is not answered by the descritpion.
You will give a rank of 5 if query perfectly matches the descritpion.
Give a rank of intemediate number from 0 to 5 if there is a partial match of key concepts in the descritpion to query.
A perfect match is when the query matches the description very closely.
if query does not entail the description give a score of 3

Please rank the following query and descritpion and just return the rank and explain why you gave the ranking

query: {query}\n 

descritpion: {descritpion}.
"""
    result=generate(model=DEFAULT_MODEL, prompt=prompt)
    return result['response']



#!/usr/bin/env python

import sys, os, logging,datetime, docx, pandas as pd
logger = logging.getLogger( "genai_utils" )
from langchain_core.documents import Document
from genai_utils import pdf_parser

# ------------------------------------------------------------------------------------------
# Following functions extract chunks
# ------------------------------------------------------------------------------------------
def extractDocx(file):
    document = docx.Document(file)

    # STEP 1: extract tables 
    tables=[]
    for table in document.tables:
        data = []
        keys = None
        for i, row in enumerate(table.rows):
            text = (cell.text for cell in row.cells)

            if i == 0:
                keys = tuple(text)
                continue
            row_data = dict(zip(keys, text))
            data.append(row_data)

        df = pd.DataFrame(data)
        
        t = tables.append(Document(metadata={"head": "table", 'source': file}, 
                                       page_content= str(df)))
        tables.append(t)

    # STEP 2: extract text 
    paras = []
    paraTexts = []
    secHeader = ""

    for para in document.paragraphs:
        if (not para.text.strip()):
            continue;
        
        if para.style.name != "Normal" and "Paragraph" not in para.style.name:
            #print("==>", para.style.name)
            if ( len(paraTexts) > 0):
                paras.append(Document(metadata={"head": secHeader, 'source': file}, 
                                       page_content="\n".join(paraTexts)))
                paraTexts = []

            secHeader = para.text 
            continue;
        
        paraTexts.append(para.text)

    docs =  tables + paras
    return docs
# ---------------------------------------------------------------------------------------
'''
OLD FUNCTION

def getChunks(file, chunk_size=8000, overlap=256 ):
    from langchain_text_splitters import RecursiveCharacterTextSplitter
    
    txt= extractText( file= file)

    split = RecursiveCharacterTextSplitter(
        chunk_size= chunk_size,
        chunk_overlap=overlap,
        length_function=len,
        add_start_index=True,
    )
    docs = []
    for txt in split.split_text(txt):
        d = Document(page_content= txt,  metadata=dict(source= file))
        docs.append(d)

    return docs
'''
# ---------------------------------------------------------------------------------------
def getchunksFromTxt(filename, chunk_overlap_ratio=0.2, chunk_size=8000):
    with open(filename, "r", encoding="utf-8", errors='ignore') as f:
        text = f.read()
        
    from langchain_text_splitters import RecursiveCharacterTextSplitter
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap= int(chunk_overlap_ratio * chunk_size),
        add_start_index=True,
    )
    texts = text_splitter.split_text(text)
    docs = []
    for t in texts:
        docs.append(Document( page_content=t,  metadata={"source": filename}))
    return docs
#-----------------------------------------------------------------------------------------    
def extractDocs(request=None, file=None, **kwargs):
    ret = f"Unknown file type {file}"

    if ( request and not file):
        for f in request.FILES.getlist('file'):
            content = f.read()
            #fileIO = io.BytesIO(content)
            file = f"/tmp/{str(f)}"
            with open(file, "wb") as f:
                f.write(content)


    if (file.endswith("doc") or file.endswith("docx") ):
        ret =  extractDocx(file)
    elif (file.endswith("txt") or file.endswith("md") ):
        ret =  getchunksFromTxt(file)
    elif (file.endswith("pdf") ):
        ret = pdf_parser.getDocsFromPDF(file)
    elif (file.endswith("xlsx") or file.endswith("xls")):
        df = pd.read_excel(file)
        ret = df.to_html()
        d = Document(page_content= ret,  metadata=dict(source= file))
        ret=[d]
    elif file.endswith("csv") :
        df = pd.read_csv(file)
        ret = df.to_html()
        d = Document(page_content= ret,  metadata=dict(source= file))
        ret=[d]
    elif file:
        with open(file, "r", encoding="utf-8", errors='ignore') as f:
            ret = f.read()
        d = Document(page_content= ret,  metadata=dict(source= file))
        ret=[d]
    else:
        ret = []
    
    return ret




from elasticsearch import Elasticsearch
from typing import Any, Dict, Optional, List

# --------------------------------------------------------------------------------
import sys, os, ollama, torch, logging,datetime, requests,re, hashlib
from ollama import Client
from mangorest.mango import webapi
    
logger = logging.getLogger( "myapp" )
device = "cpu"

if (torch.cuda.is_available() ):
    device = "cuda"
elif torch.backends.mps.is_available():
    device = "mps"

OLLAMA_HOST= 'http://127.0.0.1:11434/v1'
OPENAI_KEY = "NO KEY"

sys.path.append(os.path.expanduser("~/.django") )
if (os.path.exists(os.path.expanduser("~/.django/my_config.py"))):
    import my_config
    try:
        from my_config import OLLAMA_HOST
    except:
        pass

MODEL= "ollama3.2"
# --------------------------------------------------------------------------------
# Function to get embedding using OLLAMA API
# Generate embeddings for a given text using the OLLAMA API.
#
def get_ollama_embedding(text: str, model: str =MODEL) -> List[float]:
    url = "http://localhost:11434/api/embeddings"
    payload = {"model": model, "prompt": text}
    response = requests.post(url, json=payload)

    if response.status_code == 200:
        embedding = response.json().get("embedding", [])
        return embedding
    else:
        raise Exception(f"Failed to get embedding: {response.text}")

#query_embedding = get_ollama_embedding(query_text, model=model )



#!/usr/bin/env python 

import pdfplumber, os, base64, glob, tqdm, json
from io import BytesIO
import io, base64
from PIL import Image
from IPython.display import HTML
from genai_utils.describe_image import describe_image
from genai_utils import db_elastic

def _extractImagesFromPDF(file=None, **kwargs):
    assert file.endswith("pdf"), "Called with non PDF File!!"

    ret = {}
    txt = []
    with pdfplumber.open(file) as doc:
        for pageNumber, page in enumerate(doc.pages):
            images = page.images
            for image_index, img in enumerate(images):
                try:
                    bbox = (img['x0'], img['top'], img['x1'], img['bottom'])
                    image = page.within_bbox(bbox).to_image()
                    pil_image = image.original
                    imageRGB = pil_image.convert("RGB")
                    b = BytesIO()
                    imageRGB.save(b, format='PNG')
                    b.seek(0)
                    br= b.read()
                    b64Image = base64.b64encode(br).decode("utf-8")
                    url = "data:image/jpg;base64, " + b64Image
                    txt.append(page.extract_text())
                    #img = f"<img src='{url}' >"
                    #display (HTML(img))
                    ret[url] = 1
                except:
                    pass
    ret = [r for r in ret.keys()]
    return dict(images=ret, texts=txt)

def indexImagesFromPDF(file, savedir="/tmp/genai_utils/", verbose =0):
    ret = _extractImagesFromPDF(file)
        
    if ( savedir is None or not savedir):
        return ret, None
    files = []
    for i, img in enumerate(ret['images']):
        img1=img[img.index(",")+1:].strip()
        imgd = Image.open(io.BytesIO(base64.decodebytes(img1.encode()) ))
        
        bname = os.path.basename(file)
        sfile = f"{savedir}/{bname}__{i}.png"
        os.makedirs(savedir, exist_ok=True)
        imgd.save(sfile)
        files.append(sfile)
        print(f"Saved {sfile}")
        if ( verbose):
            display(HTML(f"<img src='{img}'> "))
            print(ret['texts'][i][0:128])
    return ret, files
    
def index_directory(directory, outf= {}, savedir="/tmp/genai_utils/", recurse=0, count=10000):
    pngs = glob.glob(os.path.join(directory, '**/*.png') , recursive=recurse)
    jpgs = glob.glob(os.path.join(directory, '**/*.jpg') , recursive=recurse)
    jpes = glob.glob(os.path.join(directory, '**/*.jpeg'), recursive=recurse)
    pdfs = glob.glob(os.path.join(directory, '**/*.pdf') , recursive=recurse)

    images= []
    for pdfFile in tqdm.tqdm(pdfs):
        print(f"Getting images from {pdfFile}")
        ret, files = indexImagesFromPDF(pdfFile)
        images.extend(files)
    
    image_paths = [*pngs, *jpgs, *jpes, *images]
    n = 0
    for image_path in tqdm.tqdm(image_paths):
        if ( n >= count):
            break;
        if image_path in outf:
            continue
        with open(image_path, 'rb') as f:
            image_data = f.read()
        try:
            description = ""
            description = describe_image(image_data)
            print(f"Indexed {image_path}: {description}")
            outf[image_path] = description
            n += 1
        except Exception as e:
            print(f"Failed to index {image_path}: {e}")
            pass
        
    return outf

def getDocs(outf):
    from langchain_core.documents import Document

    docs = []
    for k,v in outf.items():
        print(k, v[0:32])
        d = Document(page_content=v, metadata=dict(source=k) )
        docs.append(d)
    return docs

def save(outf, file="/tmp/genai_utils/images_dir.json"):
    with open(file, "wt") as f:
        f.write(json.dumps(outf))
    
def load(file="/tmp/genai_utils/images_dir.json"):
    outf = {}
    if ( os.path.exists(file)):
        with open(file, "rt") as f:
            outf = json.loads( f.read() )
    return outf

outf=load()



import os, sys

def unsetproxy():
    if 'http_proxy' in os.environ:
        del os.environ['http_proxy']
        del os.environ['https_proxy']
        del os.environ['HTTP_PROXY']
        del os.environ['HTTPS_PROXY']
    
def setproxy(proxy=""):
    os.environ['http_proxy'] = proxy
    os.environ['https_proxy']=proxy

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
try:
    from dotenv import load_dotenv, find_dotenv
    if os.path.exists(os.path.expanduser(".env")):
        load_dotenv(os.path.expanduser(".env"))
    load_dotenv(find_dotenv("./.env"))
except ImportError:
    print("dotenv not found!, skipping...")

if os.path.exists("my_config.py"):
    import my_config
    from my_config import *
else:
    home_env = os.path.expanduser("~/.django/")
    home_con = home_env+ "/my_config.py"
    if not (home_env in sys.path):
        sys.path.append(home_env)

    if os.path.exists(home_con):
        import my_config
        from my_config import *

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_from_env_or_config(key):
    v = os.environ.get(key, "")
    if (v):
        return v
    try:
        v = vars(my_config).get(key)
        return v
    except:
        return None



import re
from collections import (
    Counter,
    defaultdict,
)
from langchain_core.documents import Document

NUM_TO_DOC: dict = {0: "Document Name", 1: "Chapter", 2: "Section", 3: "SubSection", 4: "SubSubSection"}
WHITESPACE_HEADER: dict = {72: "Paragraph", 101: "Chapter", 105: "Section", 116: "SubSection", 126: "SubSubSection"}
DOC_TO_NUM: dict = {"Document Name": 0, "Chapter": 1, "Section": 2, "SubSection": 3, "SubSubSection": 4}
TABLE_SETTINGS: dict = {"vertical_strategy": "lines", "horizontal_strategy": "lines", "snap_tolerance": 2}


"""
    Converts chunks of texts and metadata into Document objects.

    Args:
        chunks (list[dict]): list of dicts of formatted chunk text and the page number of the beginning of chunk
        document_name (str): the name of the document from which the chunks were extracted

    Returns:
        list[Document]: returns list of Document objects describing the chunks and their metadata
"""
def chunks_to_doc_obj(chunks: list[dict], document_name: str) -> list[Document]:
    return [Document(page_content=c["text"], metadata={"source": document_name, "page": c["page"]}) for c in chunks]


"""
Merge records that are part of a paragraph.

Args:
    records (list[dict]): document text lines.

Returns:
    list[dict]: merged text lines that are part of a paragraph.
"""
def merge_records(records: list[dict]) -> list[dict]:
    start = 0
    end = 0
    merge_list = []

    paragraph_tol = Counter(
        round(records[i + 1]["doctop"] - records[i]["doctop"]) for i in range(len(records) - 1)
    ).most_common()[0][0]

    for i, record in enumerate(records):

        if record["header_type"] != "Paragraph":
            merge_list.append(record)

        elif i >= end:

            start = i
            end = i + 1

            while (
                end < len(records)
                and (records[end]["header_type"] == "Paragraph")
                and (round(records[end]["doctop"] - records[end - 1]["doctop"]) <= paragraph_tol)
            ):
                end += 1

            merge_list.append(
                {
                    "header_type": "Paragraph",
                    "text": " ".join((r["text"] for r in records[start:end])),
                    "doctop": records[start]["doctop"],
                    "page_number": records[start]["page_number"],
                },
            )

    return merge_list


"""
    Convert records into a dictionary with header as keys and paragraphs as values.

    Args:
        records (list[dict]): list of dictionary containing paragraph and headers.
        document_name (str): string name

    Returns:
        dict[list]: return header: paragraph with page number appended at beginning dictionary
"""
def metadata_chunks(records: list[dict], document_name: str) -> dict[list]:
    chunk_dict = defaultdict(list)

    metadata = [f"Document Name: {document_name}"]

    for rec in records:

        if rec["header_type"] != "Paragraph":

            num = DOC_TO_NUM.get(rec["header_type"])
            doc_header_type = metadata[-1].split(":")[0]
            metadata_num = DOC_TO_NUM.get(doc_header_type)

            while metadata_num >= num:
                metadata.pop()
                doc_header_type = metadata[-1].split(":")[0]
                metadata_num = DOC_TO_NUM.get(doc_header_type)

            name = f"{rec['header_type']}: {rec['text']}"

            metadata.append(name)

        else:
            chunk_dict[tuple(metadata)].append(str(rec["page_number"]) + " " + rec["text"])

    return chunk_dict


"""
Converts metadata dict into  a list of chunks. Uses max_char_limit to limit the number of characters per chunk.

Args:
    chunk_dict (dict[list]): dictionary containing headers as keys and paragraphs as values with page numbers appended at beginning of paragraphs.
    tokens (int, optional): chunk token limit . Defaults to 200.

Returns:
    list[dict]: list of dictionaries containing the chunk and page number
"""
def chunk_dict_to_list(chunk_dict: dict[list], tokens: int = 200) -> list[dict]:

    bge_tokens = 400
    tokens = bge_tokens if tokens >= bge_tokens else 200

    max_char_limit = tokens * 4

    chunks = []

    new_chunk = True
    page = None

    for k, paragraphs in chunk_dict.items():
        meta_data = "\n".join(k)
        page_num, paragraph = map(list, zip(*(s.split(" ", 1) for s in paragraphs)))
        page_num = [int(x) for x in page_num]
        small_paragraph = sum(len(i) for i in paragraph)

        if (len(meta_data) + small_paragraph) <= max_char_limit:
            small_paragraph = " ".join(paragraph)
            chunks.append({"text": meta_data + "\n\n" + small_paragraph, "page": page_num[0]})
            new_chunk = True
            continue

        for i, chunk in enumerate(paragraph):
            chunk_length = len(meta_data) + len(chunk)

            if new_chunk:
                page = page_num[i]
                new_chunk = False

            if chunk_length > max_char_limit:

                sentence_list = re.split("([?!.])", chunk)
                minichunk = meta_data + "\n\n"

                for sentence in sentence_list:

                    if len(minichunk) + len(sentence) > max_char_limit:
                        chunks.append({"text": minichunk, "page": page})
                        page = page_num[i]
                        minichunk = meta_data + "\n\n" + sentence

                    else:
                        minichunk += sentence

            else:
                chunks.append({"text": meta_data + "\n\n" + chunk, "page": page})
                new_chunk = True

    return chunks




import re, pdfplumber, os
from collections import Counter
from functools import partial
from pathlib import Path

from pdfplumber.page import Page
from pdfplumber.pdf import PDF
from pdfplumber.table import Table
from pdfplumber.utils import intersects_bbox


NUM_TO_DOC: dict = {0: "Document Name", 1: "Chapter", 2: "Section", 3: "SubSection", 4: "SubSubSection"}
WHITESPACE_HEADER: dict = {72: "Paragraph", 101: "Chapter", 105: "Section", 116: "SubSection", 126: "SubSubSection"}
DOC_TO_NUM: dict = {"Document Name": 0, "Chapter": 1, "Section": 2, "SubSection": 3, "SubSubSection": 4}
TABLE_SETTINGS: dict = {"vertical_strategy": "lines", "horizontal_strategy": "lines", "snap_tolerance": 2}
PARAGRAPH_KEY = min(WHITESPACE_HEADER.keys())


def parsePDF(file: str) -> list[dict]:
    pdf_file = Path(file)
    if not pdf_file.exists():
        raise FileExistsError("File does not exist")

    records = []

    with pdfplumber.open(pdf_file) as doc:
        for page in doc.pages:
            tables = page.find_tables(table_settings=TABLE_SETTINGS)
            nontable_page = page.filter(partial(outside_tables, tables=tables))
            char_stat_dict = character_statistics(nontable_page)
            filtered = filter_page_by_chars(nontable_page, char_stat_dict)
            records.extend(text_line_tagger(filtered))

    return sorted(records, key=lambda x: x["doctop"])


#    Character characteristics  like font name and font size.
def character_statistics(document: Page) -> dict:
    bold_counter = Counter()
    char_counter = Counter()
    size_counter = Counter()
    italic_counter = Counter()

    for char in document.chars:
        if "Bold" in char["fontname"]:
            bold_counter[char["fontname"]] += 1

        elif "Italic" in char["fontname"]:

            italic_counter[char["fontname"]] += 1
        else:
            char_counter[char["fontname"]] += 1

        size_counter[round(char["size"])] += 1

    counter_list = [bold_counter, char_counter, size_counter, italic_counter]
    final_dict = {}
    for counter in counter_list:
        if counter:
            k, v = counter.most_common()[0]
            final_dict[k] = v

    return final_dict


# PDF: returns a filtered page that does not contain table data.
def outside_tables(obj: PDF, tables: list[Table]) -> PDF:
    return not any(intersects_bbox([obj], t.bbox) for t in tables)


def filter_page_by_chars(page: Page, char_stat_dict: dict) -> Page:
    return page.filter(
        lambda x: x["object_type"] == "char" and x["fontname"] in 
                    char_stat_dict and round(x["size"]) in char_stat_dict,
    )


# Tag each line in the page appropriately
# list[dict]: records  for each text line.
def text_line_tagger(page: Page) -> list[dict]:
    page_lines = page.extract_text_lines()
    records = []

    horizontal_white_space = Counter(round(x["x0"]) for x in page_lines)

    for text in page_lines:

        if horizontal_white_space[round(text["x0"])] == 1:
            continue

        header = "Paragraph"

        if check_header(text):
            header = header_type(text)

        records.append(
            {
                "header_type": header,
                "text": text["text"],
                "doctop": text["chars"][0]["doctop"],
                "page_number": text["chars"][0]["page_number"],
            },
        )

    return records


#  Check for bold letters. => bool: return True for all bold letters.
def check_header(text: dict) -> bool:
    return all("Bold" in char["fontname"] for char in text["chars"])


"""
Assigns new header type if the text has :
1) leading number
2) white space position
3) has table or figure within its text.

These 2 criteria are mutually exclusive and are used for different cases.

Args:
    text (dict): text line dictionary
    horizontal_white_space (Counter): white space dictionary

Returns:
    str:  new header string
"""
def header_type(text: dict) -> str:

    non_headers = ["Table", "Figure"]

    if any(text["text"].startswith(non) for non in non_headers):
        return "Paragraph"

    if round(text["x0"]) == PARAGRAPH_KEY:

        return number_to_header(text["text"])

    return whitespace_to_header(text["x0"])


def number_to_header(text: str) -> str:
    """
    Assigns header type to number.

    Args:
        text (str): header and value

    Returns:
        str: header type in the form of string
    """
    nums = text.split()[0]
    total = len(re.findall(r"[0-9]+", nums))
    if total == 0:
        return "Paragraph"
    return NUM_TO_DOC.get(total, "SubSubSection")


def whitespace_to_header(x0_position: int) -> str:
    """
    Handle edge case where the tool cannot extract leading numbers.
    Assigns header type based on whitespace position.

    Args:
        x0_position (int): _description_

    Returns:
        str: _description_
    """
    min_val = float("inf")
    header = None

    for k, v in WHITESPACE_HEADER.items():
        x0_diff = abs(k - x0_position)
        if x0_diff < min_val:
            min_val = x0_diff
            header = v

    return header

# ---------------------------------------------------------------------------------------
def getDocsFromPDF(filename):
    from genai_utils.dataframe_tools import merge_records,metadata_chunks,chunk_dict_to_list, chunks_to_doc_obj
    from genai_utils import pdf_parser

    filename = os.path.expanduser(filename)
    
    record = pdf_parser.parsePDF(filename)
    # Seems like there is not data excepts figures and tables in the pdf
    if ( not record ):
        print("Hmmmm not records found in PDF file!")
        return [] 

    merged = merge_records(record)

    docName = os.path.basename(filename)
    chunk_dict = metadata_chunks(merged,docName)
    chunks = chunk_dict_to_list(chunk_dict)
    docs = chunks_to_doc_obj(chunks, docName )
    return docs



#!/usr/bin/env python

import sys, os, logging, pandas as pd
logger = logging.getLogger( "genai_utils" )
from mangorest.mango import webapi

# ------------------------------------------------------------------------------------------
# Simply extarct text rfrom PDF file
def extractTextPDF(file):
    import pdfplumber

    #from pdfplumber.page import Page
    #from pdfplumber.pdf import PDF
    #from pdfplumber.table import Table
    #from pdfplumber.utils import intersects_bbox

    text = []
    #print(f"**** OPENINF FILE ===>  {f}")
    with pdfplumber.open(file) as doc:
        for page in doc.pages:
            #lines = page.extract_text_lines()
            #txt = "\n".join([l['text'] for l in lines])
            txt = page.extract_text_simple()
            text.append(txt)

    all="\n".join(text)
    return all

# ------------------------------------------------------------------------------------------
# Simply extarct text rfrom PDF file
def extractDocx(file):
    import docx
    document = docx.Document(file)
    txts=[]
    for p in document.paragraphs:
        txts.append(p.text)

    all = "\n".join(txts)
    return all
#-----------------------------------------------------------------------------------------    
@webapi("/gpt/extractText/")
def extractText(request=None, file=None, **kwargs):
    ret = f"Unknown file type {file}"

    if ( request and not file):
        for f in request.FILES.getlist('file'):
            content = f.read()
            #fileIO = io.BytesIO(content)
            file = f"/tmp/{str(f)}"
            with open(file, "wb") as f:
                f.write(content)

    print(f"Reading file {file}")

    if (file.endswith(".doc") or file.endswith(".docx") ):
        ret =  extractDocx(file)
    elif (file.endswith(".txt") or file.endswith(".md") ):
        with open(file, "r", encoding="utf-8", errors='ignore') as f:
            ret = f.read()
    elif (file.endswith(".pdf") ):
        ret = extractTextPDF(file)
    elif (file.endswith(".xlsx") or file.endswith(".xls")):
        df = pd.read_excel(file)
        ret = df.to_html()
    elif file.endswith(".csv") :
        df = pd.read_csv(file)
        ret = df.to_html()
    elif file:
        ret = open(file, "rb").read()
    else:
        ret = ""
    
    return ret

# ---------------------------------------------------------------------------------------
def getChunks(file, chunk_size=2000, overlap=256 ):
    from langchain_text_splitters import RecursiveCharacterTextSplitter
    from langchain_core.documents import Document
    
    txt= extractText( file= file)

    split = RecursiveCharacterTextSplitter(
        chunk_size= chunk_size,
        chunk_overlap=overlap,
        length_function=len,
        add_start_index=True,
    )
    docs = []
    for txt in split.split_text(txt):
        d = Document(page_content= txt,  metadata=dict(source= file))
        docs.append(d)

    return docs
